#ifndef lint
static char sccsid[] = "$Id: read_segy_trace.c,v 1.5 2014/05/29 00:57:51 doug Exp $ ";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/file.h>
#include <string.h>
#include <fnmatch.h>

#include    "qlib2.h"

#include    "segy.h"
#include    "segy_reel.h"
#include    "procs.h"
#include    "externals.h"

void swab_segyhead (SEGYHEAD *segyhead);

/************************************************************************/
/*  read_segy_trace:							*/
/*	Read SEGY trace header and trace data.				*/
/*  return:								*/
/*	1 on success, 0 on failure.					*/
/*	Set *phdr to point to DATA_HDR.					*/
/*	Set **pdata to point to allocated data buffer.			*/
/*	Caller must free DATA_HDR and data buffer.			*/
/************************************************************************/
int read_segy_trace
   (FILE	*fp,
    SEGYBREELHEAD *bin_reel_header,
    DATA_HDR	**phdr,
    int		**pdata,
    int		segy_fmt)
{
    DATA_HDR *hdr;
    EXT_TIME et;
    SEGYDREELHEAD char_reel_header;
    SEGYHEAD segyhead;
    double drate;
    int *data = NULL;
    short int *sdata = NULL;
    int data_format, i;
    char *stn;
    
    /* Read the trace header.						*/
    if (fread(&segyhead,sizeof(SEGYHEAD),1,fp)!=1) {
	fprintf(stderr, "Error reading in a SEG-Y Binary reel header.\n");
	exit(1);
    }
    if (my_endian != SEED_BIG_ENDIAN) {
	swab_segyhead(&segyhead);
    }

    hdr = new_data_hdr();
    hdr->seq_no = 1;

    /* Set fields in DATA_HDR from SEGY headers.			*/
    switch (segy_fmt) {
      case BERKELEY_FORMAT:
	/* Berkeley station is either 8 or 6 characters,		*/
	/* beginning at first additional field or 2 bytes later.	*/
	/* Station is first 3 characters, component is rest.		*/
	stn = segyhead.station_name;
	if (stn[0] == '0' && stn[1] == '0') stn += 2;
	strncpy (hdr->station_id, stn, 3);
	strcpy (hdr->channel_id, stn + 3);
	strcpy (hdr->network_id, network);
	et.year = bin_reel_header->nyear;
	et.doy  = bin_reel_header->nday;
	et.hour = bin_reel_header->nhour;
	et.minute = bin_reel_header->nmin;
	et.second = bin_reel_header->nsec;
	et.usec = bin_reel_header->nmsec * USECS_PER_MSEC;
	dy_to_mdy (et.doy, et.year, &et.month, &et.day);
	hdr->begtime = hdr->hdrtime = ext_to_int(et);
	drate = 1000000.0 / bin_reel_header->sr;
	hdr->sample_rate = (drate >= 1.0) ? 
	    roundoff(drate) : 
	    -1. * roundoff(1.0/drate);
	hdr->sample_rate_mult = 1;
	hdr->num_samples = bin_reel_header->knsamp;
        switch (bin_reel_header->kform) {
	  case 1:   fprintf (stderr, "IBM Floating point format unsupported.\n"); exit(1);
	  case 2:   data_format = INT_32; break;
	  case 3:   data_format = INT_16; break;
	  case 4:   fprintf (stderr, "Gain range format unsupported.\n"); exit(1);
	  default:  fprintf (stderr, "Invalid data format code: %d\n", bin_reel_header->kform);
		    exit(1);
	}
	break;

      case PASSCAL_FORMAT:
	/*
	strcpy (hdr->station_id, segyhead.station_name);
	strcpy (hdr->channel_id, segyhead.channel_name);
	*/
	/* Passcal convention uses:					*/
	/*	inst_no         as station_id				*/
	/*	channel_number  as channel_id				*/
	sprintf (hdr->station_id,"%05d",segyhead.inst_no);
	sprintf (hdr->channel_id,"%03d",segyhead.channel_number);
	strcpy (hdr->network_id, network);
	et.year = segyhead.year;
	et.doy = segyhead.day;
	et.hour = segyhead.hour;
	et.minute = segyhead.minute;
	et.second = segyhead.second;
	dy_to_mdy (et.doy, et.year, &et.month, &et.day);
	et.usec = segyhead.m_secs * USECS_PER_MSEC;
	hdr->begtime = hdr->hdrtime = ext_to_int (et);
	drate = (segyhead.deltaSample == 1) ?
	    (double)1000000.0 / segyhead.samp_rate :
	    (double)1000000.0 / segyhead.deltaSample;
	hdr->sample_rate = (drate >= 1) ? 
	    roundoff(drate) : 
	    -1 * roundoff(1.0/drate);
	hdr->sample_rate_mult = 1;
	hdr->num_samples = (segyhead.sampleLength == 32767) ?
	    segyhead.num_samps :
	    segyhead.sampleLength;
	switch (segyhead.data_form) {
	  case 0:   data_format = INT_16; break;
	  case 1:   data_format = INT_32; break;
	  default:  fprintf (stderr, "Invalid data_format code: %d\n", segyhead.data_form);
		    exit(1);
	}
	break;

      default:
	fprintf (stderr, "Unknown SEGY format type\n");
	exit (1);
    }

    trim (hdr->station_id);
    trim (hdr->channel_id);
    trim (hdr->network_id);

    /* Perform optional station and channel mapping.	*/
    i = 0;
    if (map) {
	while (map[i].segy_station[0] != '\0') {
	    if (fnmatch(map[i].segy_station, hdr->station_id, 0) == 0 &&
		fnmatch(map[i].segy_channel, hdr->channel_id, 0) == 0) {
		if (strcmp(map[i].seed_station,"*") != 0) 
		    strcpy(hdr->station_id,map[i].seed_station);
		if (strcmp(map[i].seed_channel,"*") != 0) 
		    strcpy(hdr->channel_id,map[i].seed_channel);
		break;
	    }
	    i++;
	}
    }
    
    /*	Read SEGY trace data.						*/
    data = (int *)malloc(hdr->num_samples * sizeof(int));
    if (data == NULL) {
	fprintf (stderr, "Error mallocing data array for %d samples\n", hdr->num_samples);
	exit (1);
    }
    switch (data_format) {
      case INT_32:
	if (fread(data,hdr->num_samples * sizeof(int),1,fp)!=1) {
	    fprintf (stderr, "Error reading %d samples of data\n", hdr->num_samples);
	    exit(1);
	}
	*pdata = data;
	if (my_endian != SEED_BIG_ENDIAN) {
	    for (i=0; i<hdr->num_samples; i++) {
		swab4(data+i);
	    }
	}
	break;
      case INT_16:
	sdata = (short int *)malloc(hdr->num_samples * sizeof(short int));
	if (sdata == NULL) {
	    fprintf (stderr, "Error mallocing data array for %d samples\n", hdr->num_samples);
	    exit (1);
	}
	if (fread(sdata,hdr->num_samples * sizeof(short int),1,fp)!=1) {
	    fprintf (stderr, "Error reading %d samples of data\n", hdr->num_samples);
	    exit(1);
	}
	if (my_endian != SEED_BIG_ENDIAN) {
	    for (i=0; i<hdr->num_samples; i++) {
		swab2(sdata+i);
	    }
	}
	for (i=0; i<hdr->num_samples; i++) {
	    data[i] = sdata[i];
	}
	free (sdata);
	*pdata = data;
	break;
      default:
	fprintf (stderr, "Unsupported data format: %d\n", data_format);
	exit (1);
    }
    *phdr = hdr;
    return (1);
}

void swab_segyhead (SEGYHEAD *segyhead)
{
    swab4(&(segyhead->lineSeq));
    swab4(&(segyhead->reelSeq));
    swab4(&(segyhead->event_number));
    swab4(&(segyhead->channel_number));
    swab4(&(segyhead->energySourcePt));
    swab4(&(segyhead->cdpEns));
    swab4(&(segyhead->traceInEnsemble));
    swab2(&(segyhead->traceID));
    swab2(&(segyhead->vertSum));
    swab2(&(segyhead->horSum));
    swab2(&(segyhead->dataUse));
    swab4(&(segyhead->sourceToRecDist));
    swab4(&(segyhead->recElevation));
    swab4(&(segyhead->sourceSurfaceElevation));
    swab4(&(segyhead->sourceDepth));
    swab4(&(segyhead->datumElevRec));
    swab4(&(segyhead->datumElevSource));
    swab4(&(segyhead->sourceWaterDepth));
    swab4(&(segyhead->recWaterDepth));
    swab2(&(segyhead->elevationScale));
    swab2(&(segyhead->coordScale));
    swab4(&(segyhead->sourceLongOrX));
    swab4(&(segyhead->sourceLatOrY));
    swab4(&(segyhead->recLongOrX));
    swab4(&(segyhead->recLatOrY));
    swab2(&(segyhead->coordUnits));
    swab2(&(segyhead->weatheringVelocity));
    swab2(&(segyhead->subWeatheringVelocity));
    swab2(&(segyhead->sourceUpholeTime));
    swab2(&(segyhead->recUpholeTime));
    swab2(&(segyhead->sourceStaticCor));
    swab2(&(segyhead->recStaticCor));
    swab2(&(segyhead->totalStatic));
    swab2(&(segyhead->lagTimeA));
    swab2(&(segyhead->lagTimeB));
    swab2(&(segyhead->delay));
    swab2(&(segyhead->muteStart));
    swab2(&(segyhead->muteEnd));
    swab2(&(segyhead->sampleLength));
    swab2(&(segyhead->deltaSample));
    swab2(&(segyhead->gainType));
    swab2(&(segyhead->gainConst));
    swab2(&(segyhead->initialGain));
    swab2(&(segyhead->correlated));
    swab2(&(segyhead->sweepStart));
    swab2(&(segyhead->sweepEnd));
    swab2(&(segyhead->sweepLength));
    swab2(&(segyhead->sweepType));
    swab2(&(segyhead->sweepTaperAtStart));
    swab2(&(segyhead->sweepTaperAtEnd));
    swab2(&(segyhead->taperType));
    swab2(&(segyhead->aliasFreq));
    swab2(&(segyhead->aliasSlope));
    swab2(&(segyhead->notchFreq));
    swab2(&(segyhead->notchSlope));
    swab2(&(segyhead->lowCutFreq));
    swab2(&(segyhead->hiCutFreq));
    swab2(&(segyhead->lowCutSlope));
    swab2(&(segyhead->hiCutSlope));
    swab2(&(segyhead->year));
    swab2(&(segyhead->day));
    swab2(&(segyhead->hour));
    swab2(&(segyhead->minute));
    swab2(&(segyhead->second));
    swab2(&(segyhead->timeBasisCode));
    swab2(&(segyhead->traceWeightingFactor));
    swab2(&(segyhead->phoneRollPos1));
    swab2(&(segyhead->phoneFirstTrace));
    swab2(&(segyhead->phoneLastTrace));
    swab2(&(segyhead->gapSize));
    swab2(&(segyhead->taperOvertravel));
   /* PASSCAL additional fields used in trace header. */
    swab2(&(segyhead->totalStaticHi));
    swab4(&(segyhead->samp_rate));
    swab2(&(segyhead->data_form));
    swab2(&(segyhead->m_secs));
    swab2(&(segyhead->trigyear));
    swab2(&(segyhead->trigday));
    swab2(&(segyhead->trighour));
    swab2(&(segyhead->trigminute));
    swab2(&(segyhead->trigsecond));
    swab2(&(segyhead->trigmills));
    swabf(&(segyhead-> scale_fac));
    swab2(&(segyhead->inst_no));
    swab2(&(segyhead->not_to_be_used));
    swab4(&(segyhead->num_samps));
    swab4(&(segyhead->max));
    swab4(&(segyhead->min));
}
